"""A file containing useful predicates."""

from typing import List, Iterable, Callable, Any, Sequence, Union, Tuple, Iterator

from .higher_order_functions import partial, flip, reduce
from operator import lt, le, ge, gt


def identity(x):
    """Returns its argument as it is."""
    return x


def equals(a, b=None, *args):
    """
    if only ``a`` is passed, a function is returned that returns True when the ``arg`` passed to it is equal
    to ``a``; else returns True when ``a``,``b`` and ``*args`` are equal.

    with one argument

    >>> equals_one = equals(1)
    >>> equals_one(1)
    True
    >>> equals_one(2)
    False

    with two or more arguments

    >>> equals(1,1,1)
    True
    >>> equals(1,1,2)
    False

    Added in version: 0.1.0
    """
    if not b:
        return lambda x: a == x
    elif not args:
        return a == b
    else:
        # TODO ugly. refactor
        if a == b:
            for arg in args:
                if a != arg:
                    return False
            else:
                return True
        else:
            return False


def is_(a, b=None, *args):
    """
    if only ``a`` is passed, a function is returned that returns True when the ``arg`` passed is the same object
    as ``a`` ; else returns True when ``a``,``b`` and ``*args`` are.

    with one argument

    >>> d1 = {1,2,3}
    >>> d2 = {1,2,3}
    >>> is_d1 = is_(d1)
    >>> is_d1(d2)
    >>> False
    >>> d1 == d2
    >>> True

    with two or more arguments

    >>> is_(d1,d1)
    >>> True
    >>> is_(d1,d1,d2)
    >>> False


    Added in version: 0.1.0
    """
    if not b:
        return lambda x: a is x
    elif not args:
        return a is b
    else:
        # TODO ugly. refactor
        if a is b:
            for arg in args:
                if not a is arg:
                    return False
            else:
                return True
        else:
            return False


def less_than(a, b=None, *args):
    """
    if only ``a`` is passed, a function is returned that returns True when the ``arg`` passed to is less than ``a``;
    else returns True when ``a`` is less than``b`` and ``*args``.

    with one argument

    >>> less_than_one = less_than(1)
    >>> less_than_one(2)
    False
    >>> less_than_one(0)
    True


    with two or more arguments

    >>> less_than(1,2)
    >>> True
    >>> less_than(1,2,3)
    True
    >>> less_than(1,2,3,1)
    False

    Useful to use with filter

    >>> list(filter(less_than(5),range(10)))
    [0,1,2,3,4]

    Added in version: 0.1.0
    """
    if not b:
        return lambda x: x < a
    elif not args:
        return a < b
    else:

        def f(boolean, el, args):
            if not boolean:
                return False
            elif not args:
                return boolean
            else:
                return f(lt(el, args[0]), args[0], args[1:])

        return f(lt(a, b), b, args)


def less_than_eq(a, b=None, *args):
    """
    if only ``a`` is passed, a function is returned that returns True when the ``arg`` less than or equal to ``a``;
    else returns True when ``a`` is less than or equal to ``b`` and ``*args``.

    with one argument

    >>> less_than_or_eq_to_one = less_than_eq(1)
    >>> less_than_or_eq_to_one(2)
    False
    >>> less_than_or_eq_to_one(1)
    True

    with two or more arguments

    >>> less_than_eq(1,2)
    >>> True
    >>> less_than_eq(1,2,3)
    True
    >>> less_than_eq(1,2,3,1)
    True

    Useful to use with filter

    >>> list(filter(less_than_eq(5),range(10)))
    [0,1,2,3,4,5]

    Added in version: 0.1.0
    """
    if not b:
        return lambda x: x <= a
    elif not args:
        return a <= b
    else:
        # TODO ugly. refactor
        def f(boolean, el, args):
            if not boolean:
                return False
            elif not args:
                return boolean
            else:
                return f(le(el, args[0]), args[0], args[1:])

        return f(le(a, b), b, args)


def greater_than(a, b=None, *args):
    """
    if only ``a`` is passed, a function is returned that returns True when the ``arg`` passed to is greater than ``a``;
    else returns True when ``a`` is greater than``b`` and ``*args``.

    with one argument

    >>> greater_than_one = greater_than(1)
    >>> greater_than_one(2)
    True
    >>> greater_than_one(0)
    False

    with two or more arguments

    >>> greater_than(2,1)
    >>> True
    >>> greater_than(3,2,1)
    True
    >>> greater_than(3,2,1,3)
    False

    Useful to use with filter

    >>> list(filter(greater_than(5),range(10)))
    [6,7,8,9]

    Added in version: 0.1.0
    """
    if not b:
        return lambda x: x > a
    elif not args:
        return a > b
    else:
        # TODO ugly. refactor
        def f(boolean, el, args):
            if not boolean:
                return False
            elif not args:
                return boolean
            else:
                return f(gt(el, args[0]), args[0], args[1:])

        return f(gt(a, b), b, args)


def greater_than_eq(a, b=None, *args):
    """
    if only ``a`` is passed, a function is returned that returns True when the ``arg`` greater than or equal to ``a``;
    else returns True when ``a`` is greater than or equal to ``b`` and ``*args``.

    with one argument

    >>> greater_than_eq_one = greater_than_eq(1)
    >>> greater_than_eq_one(2)
    True
    >>> greater_than_eq_one(1)
    True

    with two or more arguments

    >>> greater_than_eq(2,1)
    >>> True
    >>> greater_than_eq(3,2,1)
    True
    >>> greater_than_eq(3,2,1,3)
    True

    Useful to use with filter

    >>> list(filter(greater_than_eq(5),range(10)))
    [5,6,7,8,9]

    Added in version: 0.1.0
    """
    if not b:
        return lambda x: x >= a
    elif not args:
        return a >= b
    else:
        # TODO ugly. refactor
        def f(boolean, el, args):
            if not boolean:
                return False
            elif not args:
                return boolean
            else:
                return f(ge(el, args[0]), args[0], args[1:])

        return f(ge(a, b), b, args)


def complement(
    expr: Union[bool, Callable[[Any], bool]]
) -> Union[bool, Callable[[Any], bool]]:
    """Takes in a predicate or a Boolean expression and
    returns a negated version of the predicate or expression.

    >>> complement(True)
    >>> False

    >>> def fn(el): # returns the Boolean of el
        return bool(el)
    >>> negated_fn = complement(fn)
    >>> fn(1)
    >>> True
    >>> negated_fn(1)
    >>> False

    Added in version: 0.1.0
    """

    if not isinstance(expr, Callable):
        # Wrapped around bool to handle cases like not_(1).
        # where 1 returns a boolean value of true.
        return not bool(expr)

    else:

        def negated(*args, **kwargs):
            return not expr(*args, **kwargs)

        return negated


def is_even(num: int) -> bool:
    """Returns true when num is even.

    Added in version: 0.1.0
    """
    return num % 2 == 0


def is_odd(num: int) -> bool:
    """Returns true when num is odd

    Added in version: 0.1.0
    """
    return num % 2 != 0


def is_prime(num: int) -> bool:
    """Returns true when num is prime

    Added in version: 0.1.0
    """
    if is_even(num) and num != 2:
        # You don't need to compute the whole Sieve if num is even.
        return False

    primes = [True for i in range(num + 1)]
    p = 2
    primes[0] = False
    primes[1] = False

    while p * p <= num:
        if primes[p] == True:
            # Update all multiples of p
            for i in range(p * 2, num + 1, p):
                primes[i] = False

        p += 1

    return primes[num]


def is_divisible(divident: Union[int, float], divisor: Union[int, float]) -> bool:
    """Returns true if dividend is divisible by divisor.

    Added in version: 0.1.0
    """
    return divident % divisor == 0


def is_divisible_by(divisor: Union[int, float]) -> Callable[[Union[int, float]], bool]:
    """Takes a ``divisor`` And returns a function (closure) That expects a dividend.
    returns true if it passes the divisibility test.
    for e.g.

    >>> f = is_divisible_by(5)
    >>> f(10)
    True
    >>> f(7)
    False

    This is particularly useful to use with a filter.

    >>> list(filter(is_divisible_by(5), [1,2,3,4,5,6,7,8,9,10]))
    [5, 10]

    Suppose you want to filter out numbers that are divisible by 2 or 3

    >>> list(filter(some_predicates([is_divisible_by(2), is_divisible_by(3)]), range(1, 10)))
    [2, 3, 4, 6, 8, 9, 10]

    Added in version: 0.1.0
    """
    return lambda dividend: dividend % divisor == 0


def is_numeric(entity: Any) -> bool:
    """Return True if ``entity`` Is an ``int``,  ``float``, or a ``complex``.

    Added in version: 0.1.0
    """
    return any(map(isinstance, [entity, entity, entity], [int, float, complex]))


def is_atom(entity: Any) -> bool:
    """Everything that is NOT an iterable(except strings) are considered atoms.

    >>> is_atom("plain string")
        True
    >>> is_atom(1)
        True
    >>> is_atom([1, 2])
        False

    Added in version: 0.1.0
    """
    if isinstance(entity, str):
        return True
    else:
        return not isinstance(entity, Iterable)


def contains(entity: Any, collection: Iterable) -> bool:
    """Checks whether collection contains the given entity.
    Note, won't automatically convert a tuple of keys and values to a ``dict``.

    Added in version: 0.1.0
    """
    return entity in collection


def is_empty(collection: Iterable) -> bool:
    """Returns true if the collection is empty.

    Added in version: 0.1.0
    """
    return not bool(collection)


def is_nested(collection: Iterable) -> bool:
    """returns true if a collection is nested.
    Added in version: 0.1.0
    """
    if isinstance(collection, dict):
        return any(map(lambda x: isinstance(x, dict), collection.values()))
    else:
        return any(map(complement(is_atom), collection))


def all_predicates(*predicates: Callable[[Any], bool]) -> Callable[[Any], bool]:
    """Takes a set of predicates and returns a function that takes an entity
    and checks if it satisfies all the predicates.

    >>> even_and_prime = all_predicates(is_even, is_prime)
    >>> even_and_prime(2)
    True
    >>> even_and_prime(4)
    False
    >>> even_and_prime(3)
    False

    Added in version: 0.1.0
    """
    return lambda entity: all((p(entity) for p in predicates))


def some_predicates(*predicates: Callable[[Any], bool]) -> Callable[[Any], bool]:
    """Takes a set of predicates and returns a function that takes an entity
    and checks if it satisfies some of the predicates.

    >>> even_or_prime = some_predicates(is_even, is_prime)
    >>> even_or_prime(2)
    True
    >>> even_and_prime(4)
    True
    >>> even_and_prime(3)
    True

    Added in version: 0.1.0
    """

    return lambda entity: any((p(entity) for p in predicates))



from .higher_order_functions import *
from .predicates import *
from .seq_transform import *
from .seq_traverse import *

__all__ = [
    # higher order functions
    "flip",
    "foldr",
    "comp",
    "curry",
    "trampoline",
    "threadf",
    "threadl",
    # sequence traversing functions
    "iter_",
    "reversed_",
    "first",
    "ffirst",
    "last",
    "rest",
    "second",
    "third",
    "fourth",
    "fifth",
    "butlast",
    "take",
    "drop",
    "take_while",
    "drop_while",
    "split_with",
    "count",
    "count_",
    # sequence transforming functions
    "cons",
    "conj",
    "concat",
    "argmap",
    "argzip",
    "unzip",
    "interleave",
    "flatten",
    "insert",
    "remove",
    # predicates
    "identity",
    "equals",
    "is_",
    "less_than",
    "less_than_eq",
    "greater_than",
    "greater_than_eq",
    "complement",
    "is_even",
    "is_odd",
    "is_prime",
    "is_divisible_by",
    "is_numeric",
    "is_atom",
    "contains",
    "is_empty",
    "is_nested",
    "all_predicates",
    "some_predicates",
]



"""Functions that transform sequences"""
from collections import deque
from .predicates import is_atom, is_nested, complement
from .higher_order_functions import reduce

from .seq_traverse import iter_

from typing import (
    Callable,
    Iterable,
    Generator,
    Any,
    Tuple,
    Union,
    Dict,
    Tuple,
    List,
)


def cons(arg: Any, iterable: Iterable) -> deque:
    """Returns a deque with arg as the first element.

    Adds to the left of a deque.

    >>> cons(5, [1,2,3,4])
    deque([5, 1, 2, 3, 4])

    >>> cons(3, deque([1,2]))
    deque([3, 1, 2])

    >>> cons((3, "c"), {1:"a", 2: "b"})
    deque([(3, "c"), (1, "a"), (2, "b")])

    Added in version: 0.1.0
    """

    if isinstance(iterable, dict):
        dq = deque(iterable.items())
    else:
        dq = deque(iterable)

    dq.appendleft(arg)

    return dq


def conj(iterable: Iterable, *args: Any) -> Iterable:
    """Short for conjoin, adds element to the iterable, at the appropriate end.
    Adds to the left of a deque.

    >>> conj([1,2,3,4],5)
    [1, 2, 3, 4, 5]

    >>> conj(deque([1,2]), 3,4)
    deque([4, 3, 1, 2])

    >>> conj([1,2,3,4],5,6,7,8)
    [1, 2, 3, 4, 5, 6, 7, 8]

    >>> conj([1,2,3,4],[5,6,7])
    [1, 2, 3, 4, [5, 6, 7]]

    >>> conj((1,2,3,4),5,6,7)
    (1, 2, 3, 4, 5, 6, 7)

    >>> conj(range(10), 11)
    (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11)

    >>> conj({1:"a", 2:"b"}, {3:"c"})
    {1: 'a', 2: 'b', 3: 'c'}

    Added in version: 0.1.0
    """
    if isinstance(iterable, list):

        return iterable + list(args)  # since args is a tuple

    elif isinstance(iterable, tuple):
        return iterable + args

    elif isinstance(iterable, dict):
        # Make a copy of iterable, instead of mutating it.
        conjed_dict = {**iterable}

        for d in args:
            conjed_dict.update(d)

        return conjed_dict

    elif isinstance(iterable, set):
        return {*iterable, *args}

    elif isinstance(iterable, deque):
        dq = deque(iterable)
        for element in args:
            dq.appendleft(element)
        return dq

    else:
        return tuple(iterable) + args


def concat(iterable, *args):
    """Add items to the end of the iterable.

    >>> concat([1,2,3,4],5)
    [1, 2, 3, 4, 5]
    >>> concat(deque([1,2]), 3,4)
    deque([1, 2, 3, 4])

    Added in version: 0.1.0
    """

    if isinstance(iterable, deque):
        dq = deque(iterable)

        for element in args:
            dq.append(element)

        return dq

    else:
        # Since conj Behavior for data Structures other than
        # deque are similar to concat behaviour.
        # call conj
        return conj(iterable, *args)


def argmap(functions: Iterable[Callable], args: Iterable) -> Generator:
    """Maps the same argument(s) to multiple functions.

    >>> inc = lambda x:x+1
    >>> dec = lambda x:x-1
    >>> list(argmap([inc, dec],[1]))
        [2,0]

    you can even map multiple arguments

    >>> add = lambda a,b: a+b
    >>> sub = lambda a,b:  a-b
    >>> list(argmap([add, sub], [2, 1])) # two arguments
        [3, 1]

    Added in version: 0.1.0
    """
    return (f(*args) for f in functions)


def argzip(sequence: Iterable[Callable], *args: Any) -> Generator:
    """
    Similar to zip, but instead of zipping iterables,
    It zips an argument(s) with all the values of the iterable.
    for example.

    >>> list(argzip([1,2,3,4], "number"))
    [(1, 'number'), (2, 'number'), (3, 'number'), (4, 'number')]
    >>> list(argzip([1,2,3,4], "number", "int"))
    [(1, 'number', 'int'), (2, 'number', 'int'), (3, 'number', 'int'), (4, 'number', 'int')]

    Added in version: 0.1.0
    """
    return ((a, *args) for a in sequence)


def unzip(sequence: Iterable) -> Tuple[Any]:
    """Opposite of zip. Unzip is shallow.

    >>> unzip([[1,'a'], [2,'b'], [3,'c']])
    ((1, 2, 3), ('a', 'b', 'c'))
    >>> unzip([ [1,'a','A'], [2, 'b','B'], [3,'c','C'] ])
    ((1, 2, 3), ('a', 'b', 'c'), ('A', 'B', 'C'))

    shallow nature of unzip.

    >>> unzip([ [[1,'num'],['a','str']], [[2,'num'],['b','str']] ])
    (([1, 'num'], [2, 'num']), (['a', 'str'], ['b', 'str']))

    Added in version: 0.1.0
    """

    # TODO find better name for split?
    def split(constructed, inner_lis):
        # constructed is a nested list like [[1,2,3], ['a','b','c']]
        return tuple(map(conj, constructed, inner_lis))

    def create_nested_list(sequence):
        # to be passed as an initial value to reduce
        # the number of 2nd level lists corresponds
        # to the number of elements in the inner list
        # of sequence. for e.g
        # [ [1,'a'], [2,'b], [3,'c'] ] -> ( (), () )

        return (() for i in range(len(sequence[0])))

    return reduce(split, sequence, create_nested_list(sequence))


def interleave(*seqs: Iterable) -> Tuple:
    """Similar to clojure's interleave. returns a flat sequence with the contents of iterables interleaved.

    >>> interleave([1,2,3],["a","b","c"])
    (1, 'a', 2, 'b', 3, 'c')
    >>> interleave([1,2,3],["int","int","int"], ["a","b","c"],["str","str","str" ])
    (1, 'int', 'a', 'str', 2, 'int', 'b', 'str', 3, 'int', 'c', 'str')

    Added in version: 0.1.0
    """
    return flatten(zip(*seqs))


def flatten(sequence: Iterable) -> Tuple:
    """Returns the contents of a nested sequence as a flat sequence.
    Flatten is recursive.

    >>> flatten([1,2,[3,[4],5],6,7])
    (1, 2, 3, 4, 5, 6, 7)

    Added in version: 0.1.0
    """

    def fn(initial_val, elem):
        if is_atom(elem):
            return conj(initial_val, elem)
        elif is_nested(elem):
            # recursively call flatten if elem is a nested sequence
            return conj(initial_val, *flatten(elem))
        else:
            return conj(initial_val, *elem)

    return reduce(fn, sequence, ())


def insert(element: Any, iterable: Iterable, *, key: Callable = lambda x: x) -> Tuple:
    """Inserts ``element`` right before the first element
    in the iterable that is greater than ``element``

    >>> insert(3, [1,2,4,2])
    (1,2,3,4,2)

    >>> insert((2, "b"), {1:"a", 3:"c"})
    ((1, "a"), (2, "b"), (3, "c"))

    Using the key Parameter

    >>> Person = namedtuple("Person", ("name", "age"))
    >>> person1 = Person("John", 18)
    >>> person2 = Person("Abe", 50)
    >>> person3 = Person("Cassy", 25)
    >>> insert(person3, (person1, person2), key=lambda p:p.age)
        (person1, person3, person2)
    >>> insert(person3, (person1, person2), key=lambda p:p.name)
        (person3, person1, person2)

    Added in version: 0.1.0

    """
    # TODO refactor.
    if not iterable:
        return (element,)

    it = iter_(iterable)
    accumulator = []
    elem = next(it)

    while key(elem) <= key(element):
        accumulator.append(elem)
        try:
            elem = next(it)
        except StopIteration:
            return tuple(accumulator) + (element,)

    return tuple(accumulator) + (element, elem) + tuple(it)


def remove(predicate: Callable, iterable: Iterable) -> Tuple:
    """
    Opposite of filter; Constructs an iterable of elements that falsify the predicate.

    >>> remove(lambda x: x==1, [1,1,9,1,1]
    [9]
    >>> remove(lambda x: x%2==0, range(10))
    [1,3,5,7,9] # filter would return [2,4,6,8]

    Added in version: 0.1.0
    """
    return tuple(filter(complement(predicate), iterable))



from typing import Callable, Any, Iterable, Union
from inspect import signature
from functools import partial
from .seq_traverse import iter_, reversed_


def flip(fn: Callable) -> Callable:
    """returns a function that takes in a flipped order of args.
    Usage:

    >>> f = lambda a,b : a-b
    >>> f(1,3)
    -2
    >>> f(3,1)
    2
    >>> flipped_f = flip(f)
    >>> flipped_f(3,1)
    -2
    >>> flipped_f(1,3)
    2

    Added in version: 0.1.0
    """

    def flipped(*args, **kwargs):
        return fn(*reversed_(args), **kwargs)

    return flipped


def reduce(fn, iterable, initial=None):
    """Similar to python's reduce, but can be prematurely terminated with ``reduced``.
    Works with dictionaries too.

    Usage:

    >>> # Reducing over dictionaries.
    >>> def inc_value(result, kv_pair):
            k = kv_pair[0]
            v = kv_pair[1]
            return result[k]= v+1
    >>> reduce(inc_value, {"a":1,"b":2}, {})
    {'a': 2, 'b': 3}

    >>> #premature termination with reduced
    >>> def inc_while_odd(result, element):
            if element%2==0:
                return reduced(result)
            else:
                result.append(element+1)
                return result
    >>> reduce(inc_while_odd, [1,3,5,6,7,8],[])
    [2, 4, 6]
    # increments uptil 5 (third element) and prematurely terminates.

    """
    it = iter_(iterable)

    if initial is None:
        result = next(it)
    else:
        result = initial

    for element in it:
        result = fn(result, element)

        # check if reduce needs to be prematurely terminated.
        if isinstance(result, Reduced):
            return result()  # call result to get reduced value.

    return result


def reduced(x):
    """Use with ``functionali.reduce`` to prematurely terminate ``reduce`` with the value of ``x``.

    Usage:
    >>> reduce(lambda acc, el: reduced("!"), [1,3,4])
    "!"
    # reduce is prematurely terminated and returns a value of "!"
    """
    return Reduced(x)


class Reduced:
    def __init__(self, x):
        self.x = x

    def __call__(self):
        return self.x


def foldr(fn: Callable, iterable: Iterable, initial: Any = None) -> Any:
    """Fold right. Stack safe implementation

    Added in version: 0.1.0
    """
    # Implement with reduce, for this we reverse the iterable
    # and then we flip fn since the Function signature for fn is fn(element, accumulator)
    # This is the standard function signature of any function passed to Foldright
    # the flipping is necessary because reduce expects the reducing function to have a function signature of
    # fn(accumulator, element) Which is the flipped signature of fn(element, accumulator)
    reversed_it = reversed_(iterable)

    if initial is None:
        initial = next(reversed_it)

    return reduce(flip(fn), reversed_it, initial)


def comp(*fns: Callable):
    """
    returns a composed function that takes a variable number of args,
    and applies it to ``fns`` passed from right to left.

    Added in version: 0.1.2
    """

    def composed(*args, **kwargs):
        first_func = fns[-1]  # since we are applying functions from right to left
        return foldr(lambda f, arg: f(arg), fns[:-1], first_func(*args, **kwargs))

    return composed


def curry(fn: Callable) -> Callable:
    """Returns a curried version of the function.

    >>> def fn(arg1, arg2, arg3):  # test function
        return [arg1, arg2, arg3]
    >>> curried_fn = curry(fn)
    >>> curried_fn(1)(2)(3)
        [1, 2, 3]

    Added in version: 0.1.0
    """
    num_args = len(signature(fn).parameters)

    def curried(arg, fn, num_args):
        if num_args == 1:
            # call fn if the final arg is passed
            return fn(arg)
        else:
            # call partial fn if not final arg
            f = partial(fn, arg)
            return lambda arg: curried(arg, f, num_args - 1)

    if num_args == 0:
        return fn

    return lambda arg: curried(arg, fn, num_args)


def trampoline(fn: Callable, *args: Any):
    """takes a function ``fn`` and calls if with ``*args``. if ``fn`` returns a function,
    calls the function until a function is not returned i.e. the base case is reached.
    function ``fn`` must return a function in its recursive case.
    Useful for optimizing tail recursive functions or mutual recursions.


    >>> def fact(x, curr=1, acc=1):
    >>>    if curr == x:
    >>>        return curr*acc
    >>>    else:
    >>>        return lambda: fact(x, curr+1, acc*curr)

    >>> trampoline(fact, 3) == 6
    >>> trampoline(fact, 100000000000) # does not raise RecursionError
    """
    recursive_fn = fn(*args)
    while isinstance(recursive_fn, Callable):
        recursive_fn = recursive_fn()
    return recursive_fn


def threadf(arg: Any, forms: Iterable[Union[Callable, Iterable]]) -> Any:
    """Thread first, passes ``arg`` as the first argument to the first function in ``forms``
    and passes the result as the first argument to the second form and so on.

    see also ``threadl``.

    >>> from functionali import identity
    >>> from operator import add, sub, mul
    >>> threadf(5, [identity])
    >>> 5

    >>> threadf(5, [identity, [add, 2]])
    >>> 7

    >>> threadf(5, [[sub, 2]])
    >>> 3 # threadf(5, [[sub, 2]]) -> sub(5, 2) -> 5-2 -> 3


    >>> # combining multiple functions
    >>> threadf(5, [identity, (add, 1), (sub, 1), (mul, 3)])
    15
    """

    def fn(result, form):
        if isinstance(form, Iterable):
            fn = form[0]
            args = form[1:]
            return fn(result, *args)
        else:
            return form(result)

    return reduce(fn, forms, arg)


def threadl(arg: Any, forms: Iterable[Union[Callable, Iterable]]) -> Any:
    """Thread last, passes ``arg`` as the last argument to the first function in ``forms``
    and passes the result as the last argument to the second form and so on.

    see also ``threadf``.

    >>> from functionali import identity
    >>> from operator import add, sub, mul
    >>> threadl(5, [identity])
    >>> 5

    >>> threadl(5, [identity, [add, 2]])
    >>> 7

    >>> threadl(5, [[sub, 2]])
    >>> -3 # threadl(5, [[sub, 2]]) -> sub(2, 5) -> 2-5 -> -3

    >>> # combining multiple functions
    >>> threadl(5, [identity, (add, 1), (sub, 1), (mul, 3)])
    -15
    """

    def fn(result, form):
        if isinstance(form, Iterable):
            fn = form[0]
            args = form[1:]
            return fn(*args, result)
        else:
            return form(result)

    return reduce(fn, forms, arg)



""" Functions to traverse the sequence"""
from collections import deque
from typing import (
    Iterable,
    Any,
    TypeVar,
    Iterator,
    Tuple,
    Mapping,
    Union,
    Callable,
    Iterator,
)

import sys


def iter_(iterable: Iterable) -> Iterator:
    """Returns appropriate iterator for the given iterable.
    This is mainly created because python's ``iter``
    returns an iterable of keys instead of keys and values for ``dict``.

    >>> tuple(iter_({1: "a", 2: "b", 3: "c"}))
    ((1, "a"),(2, "b"), (3, "c"))

    Added in version: 0.1.0
    """

    if isinstance(iterable, dict):
        # since iter(dict) returns a tuple of keys.
        # I want a tuple of key-value pairs
        return iter(iterable.items())

    return iter(iterable)


def reversed_(iterable: Iterable) -> Iterator:
    """Returns appropriate reversed iterator for the given iterable.
    This is mainly created because python's ``reversed``
    returns an iterable of keys instead of keys and values for ``dict``.

    >>> tuple(reversed_({1: "a", 2: "b", 3: "c"}))
    ((3, 'c'), (2, 'b'), (1, 'a'))

    Added in version: 0.1.0
    """
    # TODO drop support for python < 3.8 when support ends
    if isinstance(iterable, dict):
        # since iter(dict) returns a tuple of keys.
        # I want a tuple of key-value pairs

        if (
            sys.version_info[1] < 8
        ):  # since reversed for dicts was available only in 3.8
            return reversed([(k, v) for k, v in iterable.items()])
        else:  # pragma:no cover
            return reversed(iterable.items())

    elif not isinstance(iterable, Iterator):
        return reversed(iterable)

    else:  # if iterable is already a reversed iterator
        try:
            return reversed(iterable)
        except TypeError:
            return reversed(tuple(iterable))


def first(iterable: Iterable[Any]) -> Union[Any, None]:
    """
    Returns the first item in an iterable or ``None`` if iterable is empty.
    If iterable is a dict, returns a tuple of the First key-value pair

    >>> first([1,2,3,4,5])
    1
    >>> first({1:"a", 2:"b"})
    (1, "a")

    Added in version: 0.1.0
    """
    try:
        return next(iter_(iterable))
    except TypeError:  # Nonetype object is not iterable
        return None
    except StopIteration:
        return None


def ffirst(iterable: Iterable[Any]) -> Union[Any, None]:
    """same as ``first(first(iterable))``
    expects a nested iterable, returns None if iterable is empty

    >>> ffirst([[1,2], [3,4], [5,6]])
    1

    Added in version: 0.1.0
    """

    return first(first(iterable))


def last(iterable: Iterable[Any]) -> Union[Any, None]:
    """
    returns the last element in the iterable.

    >>> last([1,2,3,4])
    4
    >>> last({1: 'a', 2: 'b', 3: 'c'})
    (3, "c")

    Added in version: 0.1.0
    """
    if not iterable:
        return None
    else:
        it = iter_(iterable)
        return list(it)[-1]


def rest(iterable: Iterable) -> Iterator:
    """
    Returns an iterator of all but the first element in the iterable.
    If iterable is empty it returns an empty iterator.

    >>> list(rest([1,2,3,4,5]))
    [2, 3, 4, 5]

    >>> tuple(rest({1:"a", 2:"b", 3:"c"}))
    ((2,"b"), (3, "c"))

    >>> tuple(rest([]))
    ()

    Added in version: 0.1.0
    """
    try:
        it = iter_(iterable)
        next(it)  # discard value
        return it

    except StopIteration:  # If iterable is empty
        return iter([])


def second(iterable: Iterable[Any]) -> Union[Any, None]:
    """Returns the second item in iterable, or ``None`` if length is less than 2

    >>> second([1,2,3,4,5])
    2

    Added in version: 0.1.0
    """
    result = first(rest(iterable))
    if not result:  # if result is an empty iterable
        return None
    else:
        return result


def third(iterable: Iterable[Any]) -> Union[Any, None]:
    """Returns the third item in iterable, or ``None`` if length is less than 3

    >>> third([1,2,3,4,5])
    3

    Added in version: 0.1.0
    """
    result = first(rest(rest(iterable)))
    if not result:  # if result is an empty iterable
        return None
    else:
        return result


def fourth(iterable: Iterable[Any]) -> Union[Any, None]:
    """Returns the fourth item in iterable, or ``None`` if length is less than 4

    >>> fourth([1,2,3,4,5])
    4

    Added in version: 0.1.0
    """
    result = first(rest(rest(rest(iterable))))
    if not result:  # if result is an empty iterable
        return None
    else:
        return result


def fifth(iterable: Iterable[Any]) -> Union[Any, None]:
    """Returns the fifth item in iterable, or ``None`` if length is less than 5

    >>> fifth([1,2,3,4,5])
    5

    Added in version: 0.1.0
    """
    result = first(rest(rest(rest(rest(iterable)))))
    if not result:  # if result is an empty iterable
        return None
    else:
        return result


def butlast(iterable: Iterable[Any]) -> Union[Tuple[Any], None]:
    """returns an iterable of all but the last element
    in the iterable

    >>> butlast([1, 2, 3])
    (1, 2)

    Added in version: 0.1.0
    """
    # TODO Check efficiency of the operation
    # since it's iterating through it twice.
    t = tuple(iter_(iterable))[:-1]
    if t:
        return t
    else:
        return None


def take(n: int, iterable: Iterable) -> Tuple:
    """Returns the first n number of elements in iterable.
    Returns an empty tuple if iterable is empty

    >>> take(3, [1,2,3,4,5])
    (1, 2, 3)
    >>> take(2, {1: "a", 2: "b", 3: "c"})
    ((1, "a"), (2, "b"))

    Added in version: 0.1.0
    """
    it = iter_(iterable)

    accumulator = []
    i = 1
    while i <= n:
        try:
            accumulator.append(next(it))
            i += 1
        except StopIteration:
            break

    return tuple(accumulator)


def drop(n: int, iterable: Iterable) -> Tuple:
    """Returns All the Elements after the first
    n number of elements in iterable.
    Returns an empty tuple if iterable is empty

    >>> drop(3, [1,2,3,4,5])
    (4,5)
    >>> drop(2, {1: "a", 2: "b", 3: "c"})
    ((3, "c"),)

    Added in version: 0.1.0
    """

    it = iter_(iterable)

    i = 1
    while i <= n:
        try:
            next(it)  # discard values
            i += 1
        except StopIteration:
            break

    return tuple(it)


def take_while(predicate: Callable, iterable: Iterable) -> Tuple:
    """
    Constructs a iterable list by taking elements from ``iterable`` while ``predicate`` is true,
    Stop taking after the first element falsifies the predicate.


    >>> take_while(is_even, [2,4,6,7,8,9,10])
    (2,4,6) # Notice that it does not include 8 and 10

    >>> def is_even_dict(d):
            #checks if the key of dict d is even
            return d[0]%2==0
    >>> take_while(is_even_dict, {2:"a", 4:"b",5:"c"})
        ((2, "a"), (4, "b"))

    Added in version: 0.1.0
    """

    it = iter_(iterable)

    accumulator = []
    elem = next(it)
    while predicate(elem):
        accumulator.append(elem)
        elem = next(it)

    return tuple(accumulator)


def drop_while(predicate: Callable, iterable: Iterable) -> Tuple:
    """
    Drops elements from ``iterable`` while ``predicate`` is true,
    And returns a tuple of the remaining elements in ``iterable``.

    >>> drop_while(is_even, [2,4,6,7,8,9,10])
    (7,8,9, 10)

    >>> def is_even_dict(d):
            #checks if the key of dict d is even
            return d[0]%2==0
    >>> drop_while(is_even_dict, {2:"a", 4:"b",5:"c"})
        ((5, "c"),)

    Added in version: 0.1.0
    """

    it = iter_(iterable)

    elem = next(it)
    while predicate(elem):
        # discard values
        elem = next(it)

    # Since elem is the first element
    # that fails the predicate.
    # and the iterator has already moved ahead.
    # we need to include elem
    return (elem,) + tuple(it)


def split_with(predicate: Callable, iterable: Iterable) -> Tuple[Tuple, Tuple]:
    """Equivalent to ``(take_while(predicate, iterable), drop_while(predicate, iterable))``

    >>> split_with(is_even, [2, 4, 6, 7, 8, 9, 10])
    ((2, 4, 6), (7, 8, 9, 10))

    Added in version: 0.1.0
    """
    # consider implementing with reduce
    # since we are iterating through iterable twice.
    return (take_while(predicate, iterable), drop_while(predicate, iterable))


def count(iterable: Iterable) -> int:
    """
    counts the number of elements in the iterable, works with map objects, filter objets, and iterators.
    ``count`` will consume iterators, use ``count_`` if you want access to the iterators.
    Added in version: 0.1.2

    >>> count(iter([1,2,3]))
    3
    """
    if hasattr(iterable, "__len__"):
        return len(iterable)
    else:
        return len(list(iterable))


def count_(iterable: Iterable) -> Tuple[int, Iterable]:
    """
    returns a tuple of the number of elements in the iterable and the iterable itself.
    This can be used if you wish to find the length of iterators and want to consume the iterators
    later on.
    Added in version: 0.1.2
    >>> count(iter([1,2,3]))
    (3,[1,2,3])
    """
    if hasattr(iterable, "__len__"):
        return (len(iterable), iterable)
    else:
        l = list(iterable)
        return (len(l), l)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['functionali']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'functionali',
    'version': '0.2.0',
    'description': 'Functional programming tools for python. Putting the fun in functional programming 😉',
    'long_description': '# Functionali\nfunctional programming tools for python. Putting the fun in functional programming 😉\n\n[![codecov](https://codecov.io/gh/AbhinavOmprakash/functionali/branch/main/graph/badge.svg?token=75LLE4F7EY)](https://codecov.io/gh/AbhinavOmprakash/functionali)\n[![Documentation Status](https://readthedocs.org/projects/functionali/badge/?version=latest)](https://functionali.readthedocs.io/en/latest/?badge=latest)\n      \n\nFunctional programming is a fundamentally different way of solving problems, and once It clicks, it\'s pure joy after that. \nA lot of ideas in this library have been taken from Clojure and Haskell, so the credit goes to those languages. \nIf you find your favorite function missing, or find ways to improve this project, I\'d love to hear it.\n\n\n\n[Documentation](https://functionali.readthedocs.io/en/latest/)\n\n[wiki](https://functionali.readthedocs.io/en/latest/)\n\n## Installation\n```bash \npip install functionli \n```\n\n## Features\n- Polymorphic functions that operate on all sequences. Including dictionaries.\n- A bunch of commonly used predicates that enhance readability, especially when used with filters.\n\n## A note on how functionali deals with dictionaries.\nAs of now, all sequence traversing like `first` and `rest` functions treat dictionaries as a nested tuple of keys and values. \nHere\'s an example. \n\n```pycon\n>>> d = {1:"a"}\n>>> first(d)\n((1, "a"),)\n```\n\nwhile this is convenient for traversing dicts, it makes certain functions like `contains` rather awkward, since you as the user, would \nhave to explicity convert the tuple back to a dict before passing it in. \n(This might change in future)\n\n\n\n## functional programming is powerful\n \nOne of my favorite Functional implementations of an algorithm is insertion sort, here\'s how simple it \n```pycon\n>>> from functionali import foldr, insert\n>>> def insertion_sort(iterable):\n...     return foldr(insert, iterable, [])\n\n>>> insertion_sort([3,2,1])\n(1, 2, 3)\n\n# even works with dictionaries\n>>> insertion_sort({3:"c", 1: "a", 2: "b"})  \n((1, \'a\'), (2, \'b\'), (3, \'c\'))\n\n# and strings\n>>> insertion_sort("cbdasdf")\n(\'a\', \'b\', \'c\', \'d\', \'d\', \'f\', \'s\')\n>>> "".join(insertion_sort("cbdasdf"))\n\'abcddfs\' # better?\n``` \n\n## functional programs are flexible\n\nLet\'s say you wanted to find a number that was divisible by The following numbers\n`1,2,3,4,5,6,7,8,9,10` \nOne way to solve it is a bunch of nested `if` Statements and we all know how brittle those can be.\n\nLet\'s see the functional way.\n```pycon\n>>> from functionali import is_divisible_by, all_predicates\n>>> my_filter = all_predicates(*map(is_divisible_by,[1,2,3,4,5,6,7,8,9,10]))\n>>> list(filter(my_filter,range(1,10000)))\n[2520, 5040, 7560]\n```\n\nNow let\'s say the boss decides that He wants the numbers that are not \nDivisible by all these numbers\n```pycon\n>>> from functionali import complement\n>>> list(filter(complement(my_filter),range(1,10000)))\n[... 2515, 2516, 2517, 2518, 2519, 2521, 2522, 2523, 2524...] # Snipped for brevity \n# note that 2520 is not present\n```\nThe boss comes in again and he says that he Doesn\'t want numbers that are divisible by ALL these numbers but instead, ANY of these numbers `[1,2,3,4,5,6,7,8,9,10]`\n```pycon\n>>> from functionali import some_predicates\n>>> my_filter = some_predicates(*map(is_divisible_by,[1,2,3,4,5,6,7,8,9,10]))\n>>> list(filter(complement(my_filter),range(1,10000)))\n[]\n```\nThe boss realizes that all numbers are divisible by 1 and he tells you to Remove 1\n\n```pycon\n>>> my_filter = some_predicates(*map(is_divisible_by,[2,3,4,5,6,7,8,9,10]))   \n>>> list(filter(complement(my_filter),range(1,10000)))\n[1, 11, 13, 17, 19, 23, 29, 31...] # ,Snipped for brevity\n```\n\n## A few more examples\n\nFunctionali provides functions to traverse sequences(Including dictionaries), Some of the most useful ones are `first`, `rest`, `last`,`butlast`, `take`, `drop`\n```Python\nfrom functionali import first, rest, last, butlast, take, drop\n\n>>> first([1,2,3,4,5])\n    1\n\n>>> first({1:"a", 2:"b"})\n    (1, "a")\n\n>>> last([1,2,3,4])\n    4\n\n>>> list(rest([1,2,3,4,5]))\n    [2, 3, 4, 5]\n\n>>> butlast([1,2,3]) # returns all elements except the last element\n    (1,2)\n\n>>> take(3, [1,2,3,4,5])\n    (1, 2, 3)\n\n>>> drop(3, [1,2,3,4,5])\n    (4,5)\n```\n\nThere are functions that construct new sequences like `cons`, `conj`, `concat`, `insert`\n\n```Python\nfrom functionali import cons, conj, concat, insert\n>>> cons(5, [1,2,3,4])\n    deque([5, 1, 2, 3, 4]) # 5 is the \'head\' of the new list.\n\n# adds element to the iterable, at the appropriate end.\n>>> conj([1,2,3,4],5) # similar to Clojure\'s conj\n    [1, 2, 3, 4, 5]\n\n    # Adds to the left of a deque.\n>>> conj(deque([1,2]), 3,4)\n    deque([4, 3, 1, 2])\n\n# Add items to the end of the iterable.\n>>> concat([1,2,3,4],5)\n    [1, 2, 3, 4, 5]\n\n>>> concat(deque([1,2]), 3,4)\n    deque([1, 2, 3, 4])\n\n# Inserts 3 right before the first element\n# in the iterable (here:4) that is greater than 3\n>>> insert(3, [1,2,4,2])\n    (1,2,3,4,2)\n```\n\nFunctionali also comes with a number of useful predicates \n(if You can\'t find something you\'re looking for, make a pull request.)\nThese can be combined in various ways.\nfor example.\n\n```Python\nfrom functionali import is_even, is_prime, take_while\n>>> list(filter(is_even,[1,2,3,4])\n    [2,4]\n\n>>> take_while(is_prime, [2,3,5,4,6,5])) # Constructs a list while is_prime is true.\n    [2,3,5]\n\n```\n\n\n\n\n',
    'author': 'Abhinav Omprakash',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://functionali.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
